from html.parser import HTMLParser


class HTMLTextParser(HTMLParser):
    """
    Parse HTML into text
    """

    def __init__(self):
        super().__init__()
        # All text found
        self.text = ""
        # Only text outside of html tags
        self.text_outside_tags = ""
        self.level = 0

    def handle_starttag(self, tag, attrs) -> None:
        self.level += 1

    def handle_endtag(self, tag) -> None:
        self.level -= 1

    def handle_data(self, data) -> None:
        self.text += data
        if self.level == 0:
            self.text_outside_tags += data


def get_title(title: str, skip_html_tags: bool = False):
    htp = HTMLTextParser()
    htp.feed(title)
    htp.close()

    if skip_html_tags:
        return htp.text_outside_tags
    else:
        return htp.text



"""Build a PNG card for each page meant for social media."""
import hashlib
from pathlib import Path
import matplotlib
from matplotlib import pyplot as plt
import matplotlib.image as mpimg
from sphinx.util import logging

matplotlib.use("agg")

LOGGER = logging.getLogger(__name__)
HERE = Path(__file__).parent
MAX_CHAR_PAGE_TITLE = 75
MAX_CHAR_DESCRIPTION = 175

# Default configuration for this functionality
DEFAULT_SOCIAL_CONFIG = {
    "enable": True,
    "site_url": True,
    "site_title": True,
    "page_title": True,
    "description": True,
}


# Default configuration for the figure style
DEFAULT_KWARGS_FIG = {
    "enable": True,
    "site_url": True,
}


# These functions are used when creating social card objects to set MPL values.
# They must be defined here otherwise Sphinx errors when trying to pickle them.
# They are dependent on the `multiple` variable defined when the figure is created.
# Because they are depending on the figure size and renderer used to generate them.
def _set_page_title_line_width():
    return 825


def _set_description_line_width():
    return 1000


def create_social_card(
    app, config_social, site_name, page_title, description, url_text, page_path
):
    """Create a social preview card according to page metadata.

    This uses page metadata and calls a render function to generate the image.
    It also passes configuration through to the rendering function.
    If Matplotlib objects are present in the `app` environment, it reuses them.
    """

    # Add a hash to the image path based on metadata to bust caches
    # ref: https://developer.twitter.com/en/docs/twitter-for-websites/cards/guides/troubleshooting-cards#refreshing_images  # noqa
    hash = hashlib.sha1(
        (site_name + page_title + description + str(config_social)).encode()
    ).hexdigest()[:8]

    # Define the file path we'll use for this image
    path_images_relative = Path("_images/social_previews")
    filename_image = f"summary_{page_path.replace('/', '_')}_{hash}.png"

    # Absolute path used to save the image
    path_images_absolute = Path(app.builder.outdir) / path_images_relative
    path_images_absolute.mkdir(exist_ok=True, parents=True)
    path_image = path_images_absolute / filename_image

    # If the image already exists then we can just skip creating a new one.
    # This is because we hash the values of the text + images in the social card.
    # If the hash doesn't change, it means the output should be the same.
    if path_image.exists():
        return

    # These kwargs are used to generate the base figure image
    kwargs_fig = {}

    # Large image to the top right
    if config_social.get("image"):
        kwargs_fig["image"] = Path(app.builder.srcdir) / config_social.get("image")
    elif app.config.html_logo:
        kwargs_fig["image"] = Path(app.builder.srcdir) / app.config.html_logo

    # Mini image to the bottom right
    if config_social.get("image_mini"):
        kwargs_fig["image_mini"] = Path(app.builder.srcdir) / config_social.get(
            "image_mini"
        )
    else:
        kwargs_fig["image_mini"] = (
            Path(__file__).parent / "_static/sphinx-logo-shadow.png"
        )

    # Validation on the images
    for img in ["image_mini", "image"]:
        impath = kwargs_fig.get(img)
        if not impath:
            continue

        # If image is an SVG replace it with None
        if impath.suffix.lower() == ".svg":
            LOGGER.warning(f"[Social card] %s cannot be an SVG image, skipping...", img)
            kwargs_fig[img] = None

        # If image doesn't exist, throw a warning and replace with none
        if not impath.exists():
            LOGGER.warning(f"[Social card]: %s file doesn't exist, skipping...", img)
            kwargs_fig[img] = None

    # These are passed directly from the user configuration to our plotting function
    pass_through_config = ["text_color", "line_color", "background_color", "font"]
    for config in pass_through_config:
        if config_social.get(config):
            kwargs_fig[config] = config_social.get(config)

    # Generate the image and store the matplotlib objects so that we can re-use them
    if hasattr(app.env, "ogp_social_card_plt_objects"):
        plt_objects = app.env.ogp_social_card_plt_objects
    else:
        plt_objects = None
    plt_objects = render_social_card(
        path_image,
        site_name,
        page_title,
        description,
        url_text,
        plt_objects,
        kwargs_fig,
    )
    app.env.ogp_social_card_plt_objects = plt_objects

    # Path relative to build folder will be what we use for linking the URL
    path_relative_to_build = path_images_relative / filename_image
    return path_relative_to_build


def render_social_card(
    path,
    site_title=None,
    page_title=None,
    description=None,
    siteurl=None,
    plt_objects=None,
    kwargs_fig=None,
):
    """Render a social preview card with Matplotlib and write to disk."""
    # If objects is None it means this is the first time plotting.
    # Create the figure objects and return them so that we re-use them later.
    if plt_objects is None:
        (
            fig,
            txt_site_title,
            txt_page_title,
            txt_description,
            txt_url,
        ) = create_social_card_objects(**kwargs_fig)
    else:
        fig, txt_site_title, txt_page_title, txt_description, txt_url = plt_objects

    # Update the matplotlib text objects with new text from this page
    txt_site_title.set_text(site_title)
    txt_page_title.set_text(page_title)
    txt_description.set_text(description)
    txt_url.set_text(siteurl)

    # Save the image
    fig.savefig(path, facecolor=None)
    return fig, txt_site_title, txt_page_title, txt_description, txt_url


def create_social_card_objects(
    image=None,
    image_mini=None,
    page_title_color="#2f363d",
    description_color="#585e63",
    site_title_color="#585e63",
    site_url_color="#2f363d",
    background_color="white",
    line_color="#5A626B",
    font=None,
):
    """Create the Matplotlib objects for the first time."""
    # If no font specified, load the Roboto Flex font as a fallback
    if font is None:
        path_font = Path(__file__).parent / "_static/Roboto-Flex.ttf"
        roboto_font = matplotlib.font_manager.FontEntry(
            fname=str(path_font), name="Roboto Flex"
        )
        matplotlib.font_manager.fontManager.addfont(path_font)
        font = roboto_font.name

    # Because Matplotlib doesn't let you specify figures in pixels, only inches
    # This `multiple` results in a scale of about 1146px by 600px
    # Which is roughly the recommended size for OpenGraph images
    # ref: https://opengraph.xyz
    ratio = 1200 / 628
    multiple = 6
    fig = plt.figure(figsize=(ratio * multiple, multiple))
    fig.set_facecolor(background_color)

    # Text axis
    axtext = fig.add_axes((0, 0, 1, 1))

    # Image axis
    ax_x, ax_y, ax_w, ax_h = (0.65, 0.65, 0.3, 0.3)
    axim_logo = fig.add_axes((ax_x, ax_y, ax_w, ax_h), anchor="NE")

    # Image mini axis
    ax_x, ax_y, ax_w, ax_h = (0.82, 0.1, 0.1, 0.1)
    axim_mini = fig.add_axes((ax_x, ax_y, ax_w, ax_h), anchor="NE")

    # Line at the bottom axis
    axline = fig.add_axes((-0.1, -0.04, 1.2, 0.1))

    # Axes configuration
    left_margin = 0.05
    with plt.rc_context({"font.family": font}):
        # Site title
        # Smaller font, just above page title
        site_title_y_offset = 0.87
        txt_site = axtext.text(
            left_margin,
            site_title_y_offset,
            "Test site title",
            {
                "size": 24,
            },
            ha="left",
            va="top",
            wrap=True,
            c=site_title_color,
        )

        # Page title
        # A larger font for more visibility
        page_title_y_offset = 0.77

        txt_page = axtext.text(
            left_margin,
            page_title_y_offset,
            "Test page title, a bit longer to demo",
            {"size": 46, "color": "k", "fontweight": "bold"},
            ha="left",
            va="top",
            wrap=True,
            c=page_title_color,
        )

        txt_page._get_wrap_line_width = _set_page_title_line_width

        # description
        # Just below site title, smallest font and many lines.
        # Our target length is 160 characters, so it should be
        # two lines at full width with some room to spare at this length.
        description_y_offset = 0.2
        txt_description = axtext.text(
            left_margin,
            description_y_offset,
            (
                "A longer description that we use to ,"
                "show off what the descriptions look like."
            ),
            {"size": 17},
            ha="left",
            va="bottom",
            wrap=True,
            c=description_color,
        )
        txt_description._get_wrap_line_width = _set_description_line_width

        # url
        # Aligned to the left of the mini image
        url_y_axis_ofset = 0.12
        txt_url = axtext.text(
            left_margin,
            url_y_axis_ofset,
            "testurl.org",
            {"size": 22},
            ha="left",
            va="bottom",
            fontweight="bold",
            c=site_url_color,
        )

    if isinstance(image_mini, Path):
        img = mpimg.imread(image_mini)
        axim_mini.imshow(img)

    # Put the logo in the top right if it exists
    if isinstance(image, Path):
        img = mpimg.imread(image)
        yw, xw = img.shape[:2]

        # Axis is square and width is longest image axis
        longest = max([yw, xw])
        axim_logo.set_xlim([0, longest])
        axim_logo.set_ylim([longest, 0])

        # Center it on the non-long axis
        xdiff = (longest - xw) / 2
        ydiff = (longest - yw) / 2
        axim_logo.imshow(img, extent=[xdiff, xw + xdiff, yw + ydiff, ydiff])

    # Put a colored line at the bottom of the figure
    axline.hlines(0, 0, 1, lw=25, color=line_color)

    # Remove the ticks and borders from all axes for a clean look
    for ax in fig.axes:
        ax.set_axis_off()
    return fig, txt_site, txt_page, txt_description, txt_url



from typing import Any, Dict
from urllib.parse import urljoin, urlparse, urlunparse
from pathlib import Path

import docutils.nodes as nodes
from sphinx.application import Sphinx

from .descriptionparser import get_description
from .metaparser import get_meta_description
from .titleparser import get_title

try:
    import matplotlib
except ImportError:
    print("matplotlib is not installed, social cards will not be generated")
    create_social_card = None
    DEFAULT_SOCIAL_CONFIG = {}
else:
    from .socialcards import create_social_card, DEFAULT_SOCIAL_CONFIG

import os


DEFAULT_DESCRIPTION_LENGTH = 200
DEFAULT_DESCRIPTION_LENGTH_SOCIAL_CARDS = 160
DEFAULT_PAGE_LENGTH_SOCIAL_CARDS = 80

# A selection from https://www.iana.org/assignments/media-types/media-types.xhtml#image
IMAGE_MIME_TYPES = {
    "gif": "image/gif",
    "apng": "image/apng",
    "webp": "image/webp",
    "jpeg": "image/jpeg",
    "jpg": "image/jpeg",
    "png": "image/png",
    "bmp": "image/bmp",
    "heic": "image/heic",
    "heif": "image/heif",
    "tiff": "image/tiff",
}


def make_tag(property: str, content: str, type_: str = "property") -> str:
    # Parse quotation, so they won't break html tags if smart quotes are disabled
    content = content.replace('"', "&quot;")
    return f'<meta {type_}="{property}" content="{content}" />'


def get_tags(
    app: Sphinx,
    context: Dict[str, Any],
    doctree: nodes.document,
    config: Dict[str, Any],
) -> str:
    # Get field lists for per-page overrides
    fields = context["meta"]
    if fields is None:
        fields = {}

    if "ogp_disable" in fields:
        return ""

    tags = {}
    meta_tags = {}  # For non-og meta tags

    # Set length of description
    try:
        desc_len = int(
            fields.get("ogp_description_length", config["ogp_description_length"])
        )
    except ValueError:
        desc_len = DEFAULT_DESCRIPTION_LENGTH

    # Get the title and parse any html in it
    title = get_title(context["title"], skip_html_tags=False)
    title_excluding_html = get_title(context["title"], skip_html_tags=True)

    # Parse/walk doctree for metadata (tag/description)
    description = get_description(doctree, desc_len, [title, title_excluding_html])

    # title tag
    tags["og:title"] = title

    # type tag
    tags["og:type"] = config["ogp_type"]

    if os.getenv("READTHEDOCS") and not config["ogp_site_url"]:
        # readthedocs uses html_baseurl for sphinx > 1.8
        parse_result = urlparse(config["html_baseurl"])

        if config["html_baseurl"] is None:
            raise OSError("ReadTheDocs did not provide a valid canonical URL!")

        # Grab root url from canonical url
        config["ogp_site_url"] = urlunparse(
            (
                parse_result.scheme,
                parse_result.netloc,
                parse_result.path,
                "",
                "",
                "",
            )
        )

    # url tag
    # Get the URL of the specific page
    page_url = urljoin(
        config["ogp_site_url"], app.builder.get_target_uri(context["pagename"])
    )
    tags["og:url"] = page_url

    # site name tag, False disables, default to project if ogp_site_name not
    # set.
    if config["ogp_site_name"] is False:
        site_name = None
    elif config["ogp_site_name"] is None:
        site_name = config["project"]
    else:
        site_name = config["ogp_site_name"]
    if site_name:
        tags["og:site_name"] = site_name

    # description tag
    if description:
        tags["og:description"] = description

        if config["ogp_enable_meta_description"] and not get_meta_description(
            context["metatags"]
        ):
            meta_tags["description"] = description

    # image tag
    # Get basic values from config
    if "og:image" in fields:
        image_url = fields["og:image"]
        ogp_use_first_image = False
        ogp_image_alt = fields.get("og:image:alt")
        fields.pop("og:image", None)
    else:
        image_url = config["ogp_image"]
        ogp_use_first_image = config["ogp_use_first_image"]
        ogp_image_alt = fields.get("og:image:alt", config["ogp_image_alt"])

    # Decide whether to add social media card images for each page.
    # Only do this as a fallback if the user hasn't given any configuration
    # to add other images.
    config_social = DEFAULT_SOCIAL_CONFIG.copy()
    social_card_user_options = app.config.ogp_social_cards or {}
    config_social.update(social_card_user_options)
    if (
        not (image_url or ogp_use_first_image)
        and config_social.get("enable") is not False
        and create_social_card is not None
    ):
        # Description
        description_max_length = config_social.get(
            "description_max_length", DEFAULT_DESCRIPTION_LENGTH_SOCIAL_CARDS - 3
        )
        if len(description) > description_max_length:
            description = description[:description_max_length].strip() + "..."

        # Page title
        pagetitle = title
        if len(pagetitle) > DEFAULT_PAGE_LENGTH_SOCIAL_CARDS:
            pagetitle = pagetitle[:DEFAULT_PAGE_LENGTH_SOCIAL_CARDS] + "..."

        # Site URL
        site_url = config_social.get("site_url", True)
        if site_url is True:
            url_text = app.config.ogp_site_url.split("://")[-1]
        elif isinstance(site_url, str):
            url_text = site_url

        # Plot an image with the given metadata to the output path
        image_path = create_social_card(
            app,
            config_social,
            site_name,
            pagetitle,
            description,
            url_text,
            context["pagename"],
        )
        ogp_use_first_image = False

        # Alt text is taken from description unless given
        if "og:image:alt" in fields:
            ogp_image_alt = fields.get("og:image:alt")
        else:
            ogp_image_alt = description

        # Link the image in our page metadata
        # We use os.path.sep to standardize behavior acros *nix and Windows
        url = app.config.ogp_site_url.strip("/")
        image_path = str(image_path).replace(os.path.sep, "/").strip("/")
        image_url = f"{url}/{image_path}"

        # If the social card objects have been added we add special metadata for them
        # These are the dimensions *in pixels* of the card
        # They were chosen by looking at the image pixel dimensions on disk
        tags["og:image:width"] = "1146"
        tags["og:image:height"] = "600"
        meta_tags["twitter:card"] = "summary_large_image"

    fields.pop("og:image:alt", None)

    first_image = None
    if ogp_use_first_image:
        # Use the first image that is defined in the current page
        first_image = doctree.next_node(nodes.image)
        if (
            first_image
            and Path(first_image.get("uri", "")).suffix[1:].lower() in IMAGE_MIME_TYPES
        ):
            image_url = first_image["uri"]
            ogp_image_alt = first_image.get("alt", None)
        else:
            first_image = None

    if image_url:
        # temporarily disable relative image paths with field lists
        if "og:image" not in fields:
            image_url_parsed = urlparse(image_url)
            if not image_url_parsed.scheme:
                # Relative image path detected, relative to the source. Make absolute.
                if first_image:
                    root = page_url
                else:  # ogp_image is set
                    # ogp_image is defined as being relative to the site root.
                    # This workaround is to keep that functionality from breaking.
                    root = config["ogp_site_url"]

                image_url = urljoin(root, image_url_parsed.path)
            tags["og:image"] = image_url

        # Add image alt text (either provided by config or from site_name)
        if isinstance(ogp_image_alt, str):
            tags["og:image:alt"] = ogp_image_alt
        elif ogp_image_alt is None and site_name:
            tags["og:image:alt"] = site_name
        elif ogp_image_alt is None and title:
            tags["og:image:alt"] = title

    # arbitrary tags and overrides
    tags.update({k: v for k, v in fields.items() if k.startswith("og:")})

    return (
        "\n".join(
            [make_tag(p, c) for p, c in tags.items()]
            + [make_tag(p, c, "name") for p, c in meta_tags.items()]
            + config["ogp_custom_meta_tags"]
        )
        + "\n"
    )


def html_page_context(
    app: Sphinx,
    pagename: str,
    templatename: str,
    context: Dict[str, Any],
    doctree: nodes.document,
) -> None:
    if doctree:
        context["metatags"] += get_tags(app, context, doctree, app.config)


def setup(app: Sphinx) -> Dict[str, Any]:
    # ogp_site_url="" allows relative by default, even though it's not
    # officially supported by OGP.
    app.add_config_value("ogp_site_url", "", "html")
    app.add_config_value("ogp_description_length", DEFAULT_DESCRIPTION_LENGTH, "html")
    app.add_config_value("ogp_image", None, "html")
    app.add_config_value("ogp_image_alt", None, "html")
    app.add_config_value("ogp_use_first_image", False, "html")
    app.add_config_value("ogp_type", "website", "html")
    app.add_config_value("ogp_site_name", None, "html")
    app.add_config_value("ogp_social_cards", None, "html")
    app.add_config_value("ogp_custom_meta_tags", [], "html")
    app.add_config_value("ogp_enable_meta_description", True, "html")

    # Main Sphinx OpenGraph linking
    app.connect("html-page-context", html_page_context)

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }
